r"""A module to manage polarization with Jones and Stokes formalism
"""

import math

from .jonesvector import *
from .jonesmatrix import *
from .utils import *

__version__ = "1.0.4"
__author__ = "Daniel Cote <dccote@cervo.ulaval.ca>"

